#!/usr/bin/python3
import ctypes
from multiprocessing.sharedctypes import Value
import subprocess
from ctypes import *
from enum import Enum
import socket
import psutil
import json
import threading
import re
import time
import datetime
from pydbus import SystemBus
import syslog
import os
from gi.repository import GLib, Gtk, Gio
import gi
gi.require_version('Gtk', '3.0')

ppro_add = 9
ppro_del = 10
fun_ppro = 5

class kysec_ppro_type(Enum):
    PPRO_SUBS = 1
    PPRO_KEY = 2

class kysec_func_status(Enum):
	FUNC_DISABLED = 0
	FUNC_ENABLED = 1

class kysec_ppro_kernel(Structure):
    _fields_ = [
        ('path', ctypes.c_char * 4096),
        ('hash', ctypes.c_char * 65),
        ('type', ctypes.c_int)
    ]

class KeyProcessControl:
    def __init__(self):
        self.extend_so = CDLL("libkysec_extend.so")

    def kysec_process_enable(self):
        ret = self.extend_so.kysec_getstatus()
        if ret != 2:
            syslog.syslog(syslog.LOG_DEBUG, "kysec get status error")
            return -1

        ret = self.extend_so.kysec_get_func_status(c_int(fun_ppro))
        if ret == 1 or ret == 2:
            return 0

        if ret == 0:
            ret = self.extend_so.kysec_set_func_status(c_int(fun_ppro), c_int(kysec_func_status.FUNC_ENABLED.value))
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "kysec set func status error")
                return -1

            ret = self.extend_so.kysec_conf_set("kysec_ppro", c_int(kysec_func_status.FUNC_ENABLED.value))
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "kysec set conf error")
                return -1

        return 0

    def kysec_ppro_add(self, path):
        ret = self.extend_so.kysec_ppro_add(c_char_p(bytes(path, encoding = "utf8")), c_int(kysec_ppro_type.PPRO_KEY.value))
        if ret == -1:
            syslog.syslog(syslog.LOG_DEBUG, "kysec ppro add error")
            return -1

        self.extend_so.kysec_ppro_read_node_with_path.restype = POINTER(kysec_ppro_kernel)
        try:
            p = self.extend_so.kysec_ppro_read_node_with_path(c_char_p(bytes(path, encoding = "utf8")))
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "kysec ppro read node error")
            return -1

        if p:
            syslog.syslog(syslog.LOG_DEBUG, "add type = {}, path = {}, hash = {}".format(p.contents.type, str(p.contents.path, "utf8"), str(p.contents.hash, "utf8")))
            ret = self.extend_so.kysec_ppro_communicate_with_kernel(c_int(ppro_add), p.contents.hash, p.contents.type)
        else:
            syslog.syslog(syslog.LOG_DEBUG, "kysec ppro communicate error")
            return -1

        return 0

    def kysec_ppro_delete(self, path):
        self.extend_so.kysec_ppro_read_node_with_path.restype = POINTER(kysec_ppro_kernel)
        try:
            p = self.extend_so.kysec_ppro_read_node_with_path(c_char_p(bytes(path, encoding = "utf8")))
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "kysec ppro read node error")
            return -1

        if p:
            syslog.syslog(syslog.LOG_DEBUG, "del type = {}, path = {}".format(p.contents.type, str(p.contents.path, "utf8")))
            ret = self.extend_so.kysec_ppro_communicate_with_kernel(c_int(ppro_del), p.contents.hash, p.contents.type)
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "kysec ppro communicate error")
                return -1

        ret = self.extend_so.kysec_ppro_remove(c_char_p(bytes(path, encoding = "utf8")))
        if ret != 0:
            syslog.syslog(syslog.LOG_DEBUG, "kysec ppro remove error")
            return -1

        return 0

    def kysec_ppro_get(self):
        self.extend_so.kysec_ppro_read_all_data.restype = POINTER(kysec_ppro_kernel)
        count = c_int(0)
        list = self.extend_so.kysec_ppro_read_all_data(byref(count))
        return list, count.value

    def kysec_ppro_clear(self):
        list, count = self.kysec_ppro_get()
        for i in range(0, count):
            # print(list[i].type, list[i].path, list[i].hash)
            if list[i].type == kysec_ppro_type.PPRO_KEY.value:
                ret = self.kysec_ppro_delete(str(list[i].path, "utf8"))
                if ret != 0:
                    syslog.syslog(syslog.LOG_DEBUG, "kysec ppro delete error")
                    return -1
        return 0



class processMonitor:
    """
    <node>
        <interface name='com.kylin.processMonitor.startend'>
            <method name='setProcess' >
                <arg type='s' direction='in' name='list'/>
                <arg type='i' direction='out' name='ret'/>
            </method>
            <method name='getProcess' >
                <arg type='as' direction='out' name='list'/>
            </method>
        </interface>
    </node>
    """

    def __init__(self):
        syslog.syslog(syslog.LOG_DEBUG, "processMonitor init")
        try:
            self.log = SystemBus().get("org.log.sys_transmit", "/org/log/sys_transmit")
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "log service is error")
            thread_load_transmit = threading.Thread(target = self.load_transmit,args=())
            thread_load_transmit.setDaemon(True)
            thread_load_transmit.start()

        self.auditlogPath = "/var/log/audit/audit.log"
        # self.auditlogPath = "./test.log"
        self.processListPath = "/etc/processMonitorList"

        #读取本地json数据

        #开启
        logMonitorThread = threading.Thread(target=self.logMonitor, args=(self,))
        logMonitorThread.start()

        #jsonlist = '["/usr/bin/aaa","/etc/var/log/bbb","/run/lib/ccc"]'
        #list = json.loads(jsonlist)
        #print(list[2])

    @property
    def get_host_name(self):
        """获取当前用户名"""
        json_user_list = json.dumps(psutil.users())
        user_list = json.loads(json_user_list)
        if len(user_list) < 1:
            return "root"
        return user_list[0][0]

    @property
    def get_host_time(self):
        """获取本机时间
        return: cur_time
        """
        cur_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        return cur_time

    @property
    def get_host_ip(self):
        """
        查询本机ip地址
        :return: ip
        """
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(('8.8.8.8', 80))
            ip = s.getsockname()[0]
        except Exception:
            return "unknown"
        return ip

    def load_transmit(self):
        count = 0
        while count <= 5:
            time.sleep(2)
            syslog.syslog(syslog.LOG_DEBUG, "reload transmit, count = {}".format(count))
            try:
                self.log = SystemBus().get("org.log.sys_transmit", "/org/log/sys_transmit")
                break
            except Exception:
                count = count + 1
                syslog.syslog(syslog.LOG_DEBUG, "log service is error")

    def log_transmit(self, message: str):
        time = self.get_host_time
        name = self.get_host_name
        ip = self.get_host_ip
        log_message = "'time':'{}','hostname':'{}','ip':'{}','name':'processmonitor','lv':'info','message':'{}'".format(time, name, ip, message)
        syslog.syslog(syslog.LOG_DEBUG, "log_messgae = {}".format(log_message))
        try:
            ret = self.log.log_transmit('{' + log_message + '}')
            syslog.syslog(syslog.LOG_DEBUG, "sent log ret = {}".format(ret))
        except Exception:
            syslog.syslog(syslog.LOG_DEBUG, "send log message failed")

    def getProcess(self):
        key_list = []
        #清空本地数据
        controller = KeyProcessControl()
        list, count = controller.kysec_ppro_get()
        for i in range(0, count):
            # print(list[i].type, list[i].path, list[i].hash)
            if list[i].type == kysec_ppro_type.PPRO_KEY.value:
                key_list.append(str(list[i].path, "utf8"))
        syslog.syslog(syslog.LOG_DEBUG, "key_list = {}".format(key_list))
        return key_list

    def setProcess(self, list):
        list = json.loads(list)
        syslog.syslog(syslog.LOG_DEBUG, "setProcess list = {}".format(list))
        #清空本地数据
        controller = KeyProcessControl()
        ret = controller.kysec_process_enable()
        if ret != 0:
            syslog.syslog(syslog.LOG_DEBUG, "enable kysec process protect failed")
            return ret

        ret = controller.kysec_ppro_clear()
        if ret != 0:
            syslog.syslog(syslog.LOG_DEBUG, "clear kysec key process failed")
            return ret

        for path in list:
            syslog.syslog(syslog.LOG_DEBUG, "path = {}".format(path))
            #kysec设置关键进程
            ret = controller.kysec_ppro_add(path)
            if ret != 0:
                syslog.syslog(syslog.LOG_DEBUG, "add {} failed".format(path))
        return ret


    def sendInfo(self, message):
        syslog.syslog(syslog.LOG_DEBUG, "sendInfo message = {}".format(message))
        #调用dbus发送信息至小助手
        self.log_transmit(message)
        #记录发送数据用于凌晨检测，只记录当天开始但还未结束的数据
        return 0

    def dailySend(self):
        now = datetime.datetime.now()
        if now.hour == 0 and now.minute == 0 and now.second == 0:
            self.sendInfo("start")
        elif now.hour == 23 and now.minute == 59 and now.second == 59:
            self.sendInfo("end")

    @staticmethod
    def logMonitor(self):
        syslog.syslog(syslog.LOG_DEBUG, "logMonitor start")
        with open(self.auditlogPath) as f:
            f.seek(0,2)
            file_size = os.path.getsize(self.auditlogPath)
            while True:
                self.dailySend()
                new_file_size = os.path.getsize(self.auditlogPath)
                if new_file_size < file_size - 1000:
                    syslog.syslog(syslog.LOG_DEBUG, "reload audit.log")
                    f.seek(0,2)
                data = f.readlines()
                for line in data: 
                    if line != '\n':    #空行不处理
                        # print("line = {}".format(line))
                        if "{ run }"in line or "{ exit }" in line:
                            self.parseLog(line)

                file_size = new_file_size
                time.sleep(1)
                # 处理文件截断问题

    def parseLog(self, line):
        syslog.syslog(syslog.LOG_DEBUG, "parseLog")
        syslog.syslog(syslog.LOG_DEBUG, "line = {}".format(line))
        p = ".*?{ (.*?) }.*? comm=(.*?) .*?"
        info = re.findall(p, line)
        syslog.syslog(syslog.LOG_DEBUG, "info = {}".format(info))
        if len(info):
            syslog.syslog(syslog.LOG_DEBUG, "len != 0")
            message = "status:" + info[0][0] + ", comm:" + info[0][1]
            syslog.syslog(syslog.LOG_DEBUG, "message = {}".format(message))
            self.sendInfo(message)

if __name__ == "__main__":
    loop = GLib.MainLoop()
    bus = SystemBus()
    bus.publish("com.kylin.processMonitor", processMonitor())
    loop.run()
